/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Thu Oct 18 12:27:27 2001 by Jeff Dalton
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 */

package ix.icore;

import ix.iface.util.XML;	// just for debugging output
import ix.iface.util.Reporting;

import ix.util.*;

/**
 * Common class for I-X agents. <p>
 *
 * An agent is usually created by instantiating a subclass of IXAgent
 * and then calling its <code>mainStartup(String[] argv)</code> method.
 * That method will usually be inherited from IXAgent; it processes any
 * command line arguments and then calls <code>startup()</code>.
 * Thus, from the subclass's point of view, startup and initialization
 * has the following steps:
 * <ol>
 * <li>One of the subclass's constructurs
 * <li><code>mainStartup(String[] argv)</code>, which calls
 *    <ol>
 *    <li><code>processCommandLineArguments()</code>
 *    <li><code>startup()</code>
 *    </ol>
 * </ol>
 * A subclass that redefines <code>processCommandLineArguments()</code>
 * or <code>startup()</code> should normally have them call the
 * corresponding <code>super</code> method to ensure that any
 * common processing is performed.  For example, the IXAgent
 * <code>processCommandLineArguments()</code> method will set up
 * interprocess communication as specified by the argument
 * <tt>-ipc=</tt><i>methodName</i>. <p>
 *
 * To support IPC, a subclass of IXAgent must define the
 * <code>getAgentIPCName()</code> method and will typically want
 * to define
 * <ul>
 * <li><code>handleNewIssue(Issue issue)</code>
 * <li><code>handleNewReport(Report report)</code>
 * </ul>
 * or, to intervene at an earlier stage, the method that calls the
 * two above,
 * <ul>
 * <li><code>handleInput(IPC.InputMessage message)</code>
 * </ul>
 *
 * @see ix.test.SimpleIXAgent
 * @see <a href="../../../../ix/test/SimpleIXAgent.java">SimpleIXAgent.java 
 *      source code</a>
 */
public abstract class IXAgent {

    // The dreaded global variable ...
    protected static IXAgent thisAgent = null;

    /**
     * Standard constructor. <p>
     *
     * At present, this constructor makes any standard I-X changes
     * to the look and feel.
     *
     * @see ix.iface.Iface#adjustLookAndFeel()
     */
    public IXAgent() {
	if (thisAgent != null)
	    throw new Error("Attempt to create multiple IX Agents in one VM");
	else
	    thisAgent = this;

	// /\/: Theme change doesn't work if done after objects created,
	// so is done here rather than in startup().
	ix.iface.Iface.adjustLookAndFeel();
    }

    /**
     * Method called by main(String[] argv) to do initialization common
     * to all I-X agents.
     */
    public void mainStartup(String[] argv) {

	// Parse command-line arguments
	Parameters.processCommandLineArguments(argv);

	// Have agent to look at any arguments it's interested in.
	processCommandLineArguments();

	// All command-line arguments should now have been processed.
	Parameters.checkParameterUse();

	// Finish setting things up
	startup();

    }

    /**
     * Method called by the <code>mainStartup(String[] argv)</code> method
     * to perform any setup and initialization that should take place after
     * this agent's constructor has been called and command-line arguments
     * have been processed. <p>
     *
     * At present, this method does nothing, and all the work is done
     * in subclasses.
     */
    protected void startup() {
    }

    /**
     * Handles command-line arguments common to all I-X agents.
     * <pre>
     *    -debug=<i>boolean</i>
     *    -fontsize=<i>int</i>
     *    -ipc=<i>name</i>
     * </pre>
     * At present, <tt>fontsize</tt> is ignored. <p>
     * <tt>debug</tt> is used to set <code>Debug.on</code>. <p>
     * The <i>name</i> in <tt>-ipc=</tt><i>name</i> argument will be
     * interpreted by the <code>IPC.getCommunicationStrategy(String)</code>
     * method.
     *
     * @see ix.util.IPC#getCommunicationStrategy(String methodName)
     * @see ix.util.Debug#on
     */
    protected void processCommandLineArguments() {

	if (Parameters.haveParameter("fontsize"))
	    DefaultFont.setDefaultFontSize(
		Parameters.getInt("fontsize"));

	Debug.on = Parameters.getBoolean("debug", true);

	if (Parameters.haveParameter("ipc"))
	    startServer(getAgentIPCName(),
			Parameters.getParameter("ipc"));

    }

    /**
     * Returns an object that represents the agent.
     * At present, there can be only one IXAgent per VM.
     */
    public static IXAgent getAgent() {
	return thisAgent;
    }

    /**
     * Returns the object used to represent the agent as an IPC "destination".
     * This object is usually, but not necessarily, a string containing the
     * agent's name.  The details will depend on the set of communication
     * strategies that might be used.
     *
     * @see ix.util.IPC
     * @see ix.util.IPC.CommunicationStrategy
     */
    public abstract Object getAgentIPCName();

    /**
     * Set the default IPC communication strategy and set up
     * to receive messages.
     *
     * @see ix.util.IPC
     */
    protected void startServer(Object agentName, String strategyName) {
	// Create the communication strategy.
	IPC.setDefaultCommunicationStrategy(strategyName);
	// Set up to receive messages.
	IPC.setupServer(agentName, new IPC.MessageListener() {
	    public synchronized	void
		messageReceived(IPC.InputMessage message) {
		pre_handleInput(message);
	    }
	});
    }

    /**
     * Gets the message before the <code>handleInput</code> method
     * and ensures that <code>handleInput</code> is called in the
     * AWT event dispatching thread.  This is perhaps a temporary
     * measure.
     */
    protected void pre_handleInput(final IPC.InputMessage message) {
	// /\/: We have to do this somewhere, and it's awkward
	// to do it anywhere else for reports.  So for now ...
	javax.swing.SwingUtilities.invokeLater(new Runnable() {
	    public void run() {
		handleInput(message);
	    }
	});

    }

    /**
     * Handles external input in the form of an IPC.InputMessage
     * that contains an Issue or Report object.
     *
     * It calls <code>handleNewIssue</code> or <code>handleNewReport</code>
     * as appropriate.
     */
    public void handleInput(IPC.InputMessage message) {
	Object contents = message.getContents();
        if (contents instanceof Issue) {
            handleNewIssue((Issue)contents);
        }
        else if (contents instanceof Report) {
            handleNewReport((Report)contents);
        }
        else {
            displayMessage("Unexpected message contents " + contents);
        }
    }

    /**
     * Handles new issues from external sources.  Subclasses will
     * usually redefine this method.
     */
    public void handleNewIssue(Issue issue) {
        displayMessage(Reporting.issueDescription(issue));
    }

    /**
     * Handles new reports from external sources.  Subclasses will
     * usually redefine this method.
     */
    public void handleNewReport(Report report) {
        displayMessage(Reporting.reportDescription(report));
    }

    /*
     * Default message display
     */

    /**
     * A text area in a separate frame used to display information about
     * incoming messages.
     */
    protected TextAreaFrame textFrame = null;

    /**
     * Adds the specified string to the default message display.
     */
    protected void displayMessage(String message) {
	if (textFrame == null)
	    textFrame = new TextAreaFrame("Messages to " + getAgentIPCName());
	textFrame.appendLine(message);
    }

}

// Issues:
// * Should keep a record of how IXAgent.getAgent() is used.
// * If a startup() method makes the GUI visible, subclass startup()
//   methods that call super.startup() will need to know this so they can
//   ensure that subsequent GUI stuff happens in the event-handling thread.
//   It might be better to have a separate makeVisible() that happens
//   after startup(); and maybe startup() should be renamed to something
//   like completeSetup().
