/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Wed Sep 19 21:13:33 2001 by Jeff Dalton
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 */

package ix.ideel;

import java.util.*;

import ix.ideel.event.*;
import ix.icore.domain.*;
import ix.icore.domain.event.*;
import ix.icore.*;
import ix.util.*;
import ix.util.lisp.*;
import ix.util.match.*;


/**
 * An externally-driven controller for I_DEEL.
 */

public class IdeelController implements IssueManager, DomainListener {

    Domain domain;
    List controllerListeners = new LinkedList();

    LListCollector issues = new LListCollector();

    IdeelIssue otherReportsIssue = null;

    Hashtable handlerTable = new Hashtable();

    public IdeelController() {}

    public void reset() {
	issues.clear();				// Tell viewer? /\/
	otherReportsIssue = null;
    }

    public void setDomain(Domain domain) {
	this.domain = domain;
	this.domain.addDomainListener(this);
    }

    public void addControllerListener(ControllerListener listener) {
	controllerListeners.add(listener);
    }

    public synchronized void addIssue(IdeelIssue i) {
	addIssueOptions(i);
	issues.addElement(i);
	fireIssueAdded(i);
    }

    public synchronized void addIssue(Issue i) {
	Debug.assert(!(i instanceof IdeelIssue), "wrong addIssue for", i);
	addIssue(new IdeelIssue(i));
    }

    public void fireIssueAdded(IdeelIssue issue) {
	Debug.noteln("fireIssueAdded", issue);
	ControllerEvent event = new ControllerEvent(this);
	for (Iterator i = controllerListeners.iterator(); i.hasNext();) {
	    ControllerListener listener = (ControllerListener)i.next();
	    listener.issueAdded(event, issue);
	}
    }

    void addIssueOptions(IdeelIssue i) {

	// Standard options that all issues have.
	i.addOption(new IssueOption.NoActionOption());
	i.addOption(new IssueOption.ManualOption());

	// Expansion and manual primitive options from schemas
	// in the domain.
	LList schemas = domain.getMatchingSchemas
	    (i.pattern, domain.getAllSchemas());
	for (Enumeration e = schemas.elements();
	     e.hasMoreElements();) {
	    Schema s = (Schema)e.nextElement();
	    if (!s.nodes.isNull()) {
		i.addOption(new ExpandOption(i, s));
	    }
	    else {
		// When there are no subnodes, the schema name still
		// specifies a way of doing it manually.
		i.addOption(new IssueOption.ManualOption(s.name));
	    }
	}

	// Options from I_DEEL's issue handlers.
	// At present, there is at most one.
	IdeelIssueHandler h = getIssueHandler(i);
	if (h != null) {
	    // The handler decides what options to add, but usually
	    // it will be a new AutomaticOption(i, h).
	    h.addIssueOptions(i);
	}

    }

    public synchronized void handleIssue(IdeelIssue i, IssueOption opt) {
	Debug.assert(issues.contents().find(i), "Stray issue", i);
	Debug.assert(i.hasOption(opt), "Stray issue option", opt);

	opt.handleIssue(i);
    }

    public synchronized void addIssueHandler(Object verb, IssueHandler h) {
	// At some point, this should see if any issues should get
	// an additional option, in case a handler is added after
	// some issues exist.
	String key = verb.toString();
	if (handlerTable.get(key) != null)
	    Debug.warn("Replacing handler for verb " + Util.quote(key));
	handlerTable.put(key, h);
    }

    IdeelIssueHandler getIssueHandler(IdeelIssue i) {
	return (IdeelIssueHandler)handlerTable.get(i.verb.toString());
    }

    public void expandIssue(IdeelIssue i, Object instructions) {
	// This is really a kind of issue handler.  /\/
	Schema s = (Schema)instructions;
	// Avoid null pointer exceptions in debugging code:  /\/
	if (s.name == null)
	    s.name = "temporary partial schema";
	// Note that we don't add the option to the issue.  /\/
	// Instead we handle "directly".
	IssueOption opt = new ManualExpandOption(i, s);
	i.addOption(opt);
	opt.handleIssue(i);
	fireIssueHandled(i, opt);
    }
    
    public void fireIssueHandled(IdeelIssue issue, IssueHandler handler) {
	Debug.noteln("fireIssueHandled " + issue + " " +
		     handler.getActionDescription());
	ControllerEvent event = new ControllerEvent(this);
	for (Iterator i = controllerListeners.iterator(); i.hasNext();) {
	    ControllerListener listener = (ControllerListener)i.next();
	    listener.issueHandled(event, issue, handler);
	}
    }


    /*
     * Handle new report
     */

    public synchronized void newReport(Report report) {
	for (Enumeration e = issues.elements(); e.hasMoreElements();) {
	    IdeelIssue i = (IdeelIssue)e.nextElement();
	    if (!(i instanceof OtherReportsIssue) && i.wantsReport(report)) {
		i.addReport(report);
		return;
	    }
	}
	if (otherReportsIssue == null) {
	    otherReportsIssue = new OtherReportsIssue("Note other reports");
	    addIssue(otherReportsIssue);
	}
	otherReportsIssue.addReport(report);
    }

    public static class OtherReportsIssue extends IdeelIssue {

	public OtherReportsIssue(String text) {
	    super(text);
	    setStatus(STATUS_EXECUTING);
	}

	boolean wantsReport(Report report) {
	    Debug.noteln("Other reports issue asked if it wants", report);
	    return true;
	}

    }



    /*
     * DomainListener methods
     */

    public void schemaAdded(SchemaEvent event) {
	Schema s = event.getSchema();
	Debug.noteln("Controller noting new schema", s);

	// See if any issues now have a new way to expand.
	for (Enumeration e = issues.elements(); e.hasMoreElements();) {
	    IdeelIssue i = (IdeelIssue)e.nextElement();
	    MatchEnv env = SimpleMatcher.match(s.pattern, i.pattern);
	    if (env == null) {
		continue;
	    }
	    else if (s.nodes.isNull()) {
		Debug.noteln("New manual primitive for", i);
		i.addOption(new IssueOption.ManualOption(s.name));
	    }
	    else {
		Debug.noteln("New expansion for", i);
		i.addOption(new ExpandOption(i, s.instantiate(env)));
	    }
	}
    }

    /*
     * Issue options
     */

    /**
     * Handles an issue by expanding it into subissues
     * as specified by a schema.
     */
    class ExpandOption extends IssueOption {

	Issue issue;
	Schema schema;

	ExpandOption(Issue issue, Schema schema) {
	    this.issue = issue;
	    this.schema = schema;
	    this.shortDescription = "Expand using " + schema.name;
	}

	public void handleIssue(Issue issue) {
	    // That an option's been selecte implies that execution
	    // has begun.  /\/: But subnodes must be put in first
	    // or they will get the wrong status as they are added
	    // (because parent will be executing and orderings won't
	    // yet be in so it will look like all preconditions have
	    // been met).
	    do_expansion((IdeelIssue)issue, schema);
	    Debug.assert(issue.getStatus() == STATUS_POSSIBLE);
	    issue.setStatus(STATUS_EXECUTING);
	}

	protected void do_expansion(IdeelIssue issue, Schema schema) {
	    Debug.noteln("Expanding using", schema.toLList());
	    issue.expandOneLevel(schema);
	    for (Iterator i = issue.children.iterator(); i.hasNext();) {
		IdeelIssue child = (IdeelIssue)i.next();
		IdeelController.this.addIssue(child);
	    }
	}

    }

    /**
     * Handles an issue by expanding it into subissues as specified
     * explicitly by the user.
     */
    class ManualExpandOption extends ExpandOption {

	ManualExpandOption(Issue issue, Schema schema) {
	    super(issue, schema);
	    this.shortDescription = "Expand as below";
	}

	public void handleIssue(Issue issue) {
	    do_expansion((IdeelIssue)issue, schema);
	    if (issue.getStatus() == STATUS_POSSIBLE)
		issue.setStatus(STATUS_EXECUTING);
	}

    }


    /**
     * Handles an issue by invoking one of the agent-level
     * issue-handlers.
     */
    static class AutomaticOption extends IssueOption {

	Issue issue;
	IssueHandler handler;

	AutomaticOption(Issue i, IssueHandler h) {
	     this.issue = i;
	     this.handler = h;
	     this.shortDescription = "Do it using "
		 + h.getActionDescription();
	}

	public void handleIssue(Issue i) {
	    Debug.noteln("Giving issue to handler", handler);
	    handler.handleIssue(i);
	    // N.B. It is up to that handler to do:
	    //   i.setStatus(STATUS_COMPLETE);
	    // or whatever other status changes are required.
	}

    }

}
