/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Mon Jun 18 01:48:53 2001 by Jeff Dalton
 * Copyright: (c) 2000, AIAI, University of Edinburgh
 */

package ix.iface.domain;

import java.io.*;
import java.util.*;

import ix.icore.domain.*;
import ix.util.*;
import ix.util.lisp.*;
import ix.util.match.*;

// Imports for using Xerces parser
// import org.apache.xerces.parsers.DOMParser;
// import org.w3c.dom.Document;
// import org.xml.sax.SAXException;
// import java.io.IOException;

// Imports for using JDOM
// import java.io.IOException;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;


/**
 * A parser for domains described in a simple XML syntax.
 */
public class XMLTF_Parser extends DomainParser {

    String filename;

    public XMLTF_Parser(String filename) throws FileNotFoundException {
	this.filename = filename;
    }

    public XMLTF_Parser(File file) throws FileNotFoundException {
	this(file.getPath());
    }

    public Domain readDomain() {
	return readDomain(new Domain());
    }

    public Domain readDomain(Domain domain) {
	return readDomainJDOM(domain);
    }

    //
    // Parsing with JDOM
    //

    protected String SAXDriverClass = "org.apache.xerces.parsers.SAXParser";

    public Domain readDomainJDOM(Domain domain) {

	try {
	    SAXBuilder builder = new SAXBuilder(SAXDriverClass);
	    Document doc = builder.build(new File(filename));

	    try { fillInDomain(domain, doc); }
	    catch (Error err) {
		Debug.noteln("Error processing domain", err);
	    }

	}
	catch (JDOMException jde) {
	    Debug.noteException(jde);
	}
	return domain;
    }

    protected void fillInDomain(Domain domain, Document doc) {

	Element root = doc.getRootElement();
	requireName("domain", root.getName(), root);

        List topLevelElts = root.getChildren();
	Element activitiesElt = (Element)topLevelElts.get(0);
	requireName("activities", activitiesElt.getName(), activitiesElt);

	for (Iterator i = activitiesElt.getChildren().iterator(); 
	     i.hasNext();) {
	    domain.addSchema(makeSchema((Element)i.next()));
	}

    }

    protected Schema makeSchema(Element actElt) {
	// It must be an "activity" element.
	requireName("activity", actElt.getName(), actElt);

	// Get top-level subinformation from attributes and elements
	String name = actElt.getAttributeValue("name");
	Element expandsElt = actElt.getChild("expands");
	Element nodesElt = actElt.getChild("nodes");
	Element constraintsElt = actElt.getChild("constraints");

	// Convert to instances of desired classes
	LList pattern = (LList)Lisp.readFromString(expandsElt.getText());
	LList nodes = nodesElt != null ? makeNodes(nodesElt) : Lisp.NIL;
	LList orderings = constraintsElt != null
	    ? makeOrderings(constraintsElt)
	    : Lisp.NIL;

	// Make the schema
	Schema s = new Schema(name, pattern);
	s.nodes = nodes;
	s.orderings = orderings;

	Debug.noteln("Schema from XML:", s.toLList());

	return s;

    }

    protected LList makeNodes(Element nodesElt) {
	requireName("nodes", nodesElt.getName(), nodesElt);
	LListCollector nodes = new LListCollector();
	for (Iterator i = nodesElt.getChildren().iterator(); i.hasNext();) {
	    nodes.add(makeNode((Element)i.next()));
	}
	return nodes.contents();
    }

    protected LList makeNode(Element nodeElt) {
	requireName("node", nodeElt.getName(), nodeElt);
	String id = nodeElt.getAttributeValue("id");
	LList pattern = (LList)Lisp.readFromString(nodeElt.getText());
	Long n = (Long)Lisp.readFromString(id);
	return Lisp.list(n, pattern);
    }

    protected LList makeOrderings(Element ordsElt) {
	requireName("constraints", ordsElt.getName(), ordsElt);
	LListCollector orderings = new LListCollector();
	for (Iterator i = ordsElt.getChildren().iterator(); i.hasNext();) {
	    orderings.add(makeOrdering((Element)i.next()));
	}
	return orderings.contents();
    }

    /* package */ static Object temporalSyntax = 
	Lisp.readFromString("(temporal before ?from ?to)");

    /* package */ static final Symbol
	Q_FROM = Symbol.intern("?from"),
	Q_TO   = Symbol.intern("?to");

    protected LList makeOrdering(Element ordElt) {
	requireName("constraint", ordElt.getName(), ordElt);
	LList ordSpec = (LList)Lisp.readFromString(ordElt.getText());
	MatchEnv env = SimpleMatcher.mustMatch(temporalSyntax, ordSpec);
	Long from = (Long)env.get(Q_FROM);
	Long to = (Long)env.get(Q_TO);
	return Lisp.list(from, to);
    }

    protected void requireName(String want, String have, Object source) {
	if (!have.equals(want)) {
	    throw new Error("Expected " + Util.quote(want)
			    + " but found " + Util.quote(have)
			    + " in " + source);
	}
    }



    //
    // Plain Xerces/DOM parsing - incomplete but kept for future reference.
    //

    /*

    public Domain readDomainXerces(Domain domain) {
	DOMParser parser = new DOMParser();
	try {
	    parser.setFeature(
	        "http://apache.org/xml/features/dom/defer-node-expansion",
		false);
	    parser.parse(filename);
	} catch (SAXException se) {
	    se.printStackTrace();
	} catch (IOException ioe) {
	    ioe.printStackTrace();
	}
	Document document = parser.getDocument();

	// /\/: Would have to process the DOM here.
	
	return domain;
    }

    */

}
