/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Wed Jul 18 03:55:03 2001 by Jeff Dalton
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 */

package ix.util;

import java.util.*;
import java.io.*;
import java.net.*;

import ix.iface.util.XML;

/**
 * Support for interprocess communication in a framework that allows
 * different communication strategies to be used in a uniform way.
 * A default communication strategy that sends serialized objects
 * via sockets is included. <p>
 *
 * The framework specifies certain entities only a in very general
 * way.  Communication strategies that are meant to be interchangeable
 * must therefore follow shared conventions - not enforced by the
 * framework - chiefly concerning the objects used to represent
 * "destinations" and implementations of the InputMessage interface. <p>
 *
 * Note that a "destination" need not contain all the information
 * needed for communication with the corresponding agent.  Instead,
 * the destination may be merely a name, perhaps as a String, with
 * the rest of the information stored elsewhere.
 */

public class IPC {

    private IPC() { }		// can't instantiate

    /**
     * The communication strategy used by the static methods of this
     * class that do not have a communication strategy as a parameter.
     */
    static CommunicationStrategy defaultCommunicationStrategy
	= new SimpleIXCommunicationStrategy() {};

    /**
     * Sets the communication strategy used when none is explicitly
     * specified.
     */
    public static void setDefaultCommunicationStrategy
	    (CommunicationStrategy s) {
	Debug.noteln("Setting default communication strategy to", s);
	defaultCommunicationStrategy = s;
    }

    /**
     * Sets the communication strategy used when none is explicitly
     * specified.  The <code>strategyName</code> is interpreted by
     * the <code>getCommunicationStrategy(String)</code> method.
     *
     * @see #getCommunicationStrategy(String)
     */
    public static void setDefaultCommunicationStrategy(String strategyName) {
	setDefaultCommunicationStrategy
	    (getCommunicationStrategy(strategyName));
    }

    /**
     * Returns a CommunicationStrategy based on a name that is either
     * a special abbreviation or the name of a class.  The algorithm
     * is as follows:
     * <ol>
     * <li>if <i>name</i> is <tt>""</tt>,
     *     use <code>SimpleIXCommunicationStrategy</code>;
     * <li>else if <i>name</i> is <tt>"xml"</tt>,
     *     use <code>SimpleIXXMLCommunicationStrategy</code>;
     * <li>else if a class named <i>name</i> exists,
     *     use that class;
     * <li>else if a class named 
     *        <tt>ix.</tt><i>name.Name</i><tt>CommunicationStrategy</tt>
     *        exists, where <i>Name</i> is <i>name</i> with its first
     *        character changed to upper case,
     *     use that class;
     * <li>else throw an <code>IllegalArgumentException</code>.
     * </ol>
     *
     * @see IPC.SimpleIXCommunicationStrategy
     * @see IPC.SimpleIXXMLCommunicationStrategy
     * @see #setDefaultCommunicationStrategy(String strategyName)
     */
    public static CommunicationStrategy
	getCommunicationStrategy(String strategyName) {

	// Try special case abbreviations
	if (strategyName.equals(""))
	    return new SimpleIXCommunicationStrategy();
	else if (strategyName.equals("xml"))
	    return new SimpleIXXMLCommunicationStrategy();

	Class c;
	if (// First try name as-is
	    (c = findClassForName(strategyName)) != null
	    ||
	    // Then as ix.name.NameCommunicationStrategy
	    (c = findClassForName
	             ("ix." + strategyName + "."
		      + Util.capitalizeString(strategyName)
		      + "CommunicationStrategy")) != null)
	    // We seem to have a class ...
	    try {
		return (CommunicationStrategy)c.newInstance();
	    }
	    catch (Exception e) {
		Debug.noteException(e);
		throw new IllegalArgumentException
		    ("Can't make communication strategy from " + strategyName +
		     " because " + e);
	    }
	else
	    throw new IllegalArgumentException
		("Can't find communication strategy class specified by " +
		 Util.quote(strategyName));
    }

    private static Class findClassForName(String name) {
	Debug.noteln("Looking for class", name);
	try { return Class.forName(name); }
	catch (ClassNotFoundException e) { return null; }
    }

    /**
     * Sends an object to the specified destination using the default
     * communication strategy.  The agents who are sent these objects
     * must be using a compatible strategy.  The recipient should in
     * effect get a copy of the object that was sent, but the exact
     * relationship between the two objects depends on the communication
     * strategy.
     */
    public static void sendObject(Object destination, Object contents) {
	defaultCommunicationStrategy.sendObject(destination, contents);
    }

    /**
     * Like sendObject but catches any IPC.IPCException.  It can be used
     * in cases where you don't want to detect a failure and don't want
     * an exception to be thrown any further.
     *
     * @see IPC.IPCException
     * @see #sendObject(Object, Object)
     */
    public static void exceptionlessSend(Object destination, Object contents) {
	try { sendObject(destination, contents); }
	catch (IPC.IPCException e) {
	    Debug.noteln("Send failure: " + e.getMessage());
	}
    }

    /**
     * Does what is required for this agent to receive messages sent
     * using the default communication strategy.  Here the "destination"
     * is the agent in which <tt>setupServer</tt> was called, not a
     * remote agent.  The (remote) agents sending messages to the server
     * must be using a compatible communication strategy.  Note that
     * <tt>setupServer</tt> does not assume that all messages are
     * "sending an object" in the sense of the <tt>sendObject</tt> method.
     */
    public static void setupServer(Object destination,
				   MessageListener listener) {
	defaultCommunicationStrategy.setupServer(destination, listener);
    }


    /**
     * An object that determines how various IPC operations are
     * performed.
     */
    public static interface CommunicationStrategy {
	public void sendObject(Object destination, Object contents);
	public void setupServer(Object destination,
				MessageListener listener);
    }

    /**
     * An object that represents an incoming message.  The InputMessage
     * provides a place for information that is not part of the message
     * contents but may be used for debugging or for operations such as
     * sending replies.  In later versions, this interface may include
     * methods for accessing such information.
     */
    public static interface InputMessage {
	public Object getContents();
    }

    /**
     * A minimal implementation of InputMessage.
     */
    public static class BasicInputMessage implements InputMessage {
	protected Object contents;
	public BasicInputMessage(Object contents) {
	    this.contents = contents;
	}
	public Object getContents() { return contents; }
    }


    /**
     * An object that is notified when a message is received.
     */
    public static interface MessageListener {
	public void messageReceived(InputMessage message);
    }


    /**
     * A CommunicationStrategy that provides "connections" analogous
     * to sockets and a visible mapping from destination names to the
     * data needed to establish a connection.
     */
    public static interface SocketlikeCommunicationStrategy 
      extends CommunicationStrategy {
	public Object getDestinationData(Object destination);
	public void setDestinationData(Object destination, Object data);
	public Connection connectTo(Object destination);
	public Connection getConnection(Object destination);
    }


    /**
     * An object that can send and receive.
     */
    public static interface Connection {
	public void send(Object contents);
	public Object receive();
    }


    /**
     * A mapping from destination names to the data needed to establish
     * connections with the corresponding agents.
     */
    public static interface DestinationTable {
	public Object get(Object destination);
	public Object put(Object destination, Object data);
    }

    /**
     * A HashMap implementation of the DestinationTable interface.
     */
    public static class BasicDestinationTable extends HashMap
	implements DestinationTable {
	public BasicDestinationTable() { }
    }

    /**
     * A DestinationTable that provides default host and port assignments
     * for standard I-X agents, suitable for use with an instance of
     * ObjectStreamCommunicationStrategy. <p>
     *
     * <table border=1>
     * <tr> <th>Name</th>   <th>Host</th>      <th>Port</th></tr>
     * <tr> <td>IDEEL</td> <td>localhost</td> <td>5040</td></tr>
     * <tr> <td>ILEED</td> <td>localhost</td> <td>5050</td></tr>
     * <tr> <td>ITEST</td> <td>localhost</td> <td>5060</td></tr>
     * </table>
     */
    public static class SimpleIXDestinationTable
	extends BasicDestinationTable {
	public SimpleIXDestinationTable() { }
	{
	    put("IDEEL", new ServiceAddress("localhost", 5040));
	    put("ILEED", new ServiceAddress("localhost", 5050));
	    put("ITEST", new ServiceAddress("localhost", 5060));
	}
    }


    /**
     * An ObjectStream communication strategy that provides default
     * host and port assignments for standard I-X agents.
     *
     * @see IPC.SimpleIXDestinationTable
     */
    public static class SimpleIXCommunicationStrategy
	extends ObjectStreamCommunicationStrategy {
	public SimpleIXCommunicationStrategy() { 
	    this.destinationTable = new SimpleIXDestinationTable();
	}
    }

    /**
     * An ObjectStream communication strategy that provides default
     * host and port assignments for standard I-X agents and encodes
     * the message contents in XML rather than serializing.
     *
     * @see IPC.SimpleIXDestinationTable
     */
    public static class SimpleIXXMLCommunicationStrategy
	extends XMLObjectStreamCommunicationStrategy {
	public SimpleIXXMLCommunicationStrategy() { 
	    this.destinationTable = new SimpleIXDestinationTable();
	}
    }


    /**
     * A version of ObjectStreamCommunicationStrategy that encodes
     * the message contents in XML rather than serializing.
     */
    public static class XMLObjectStreamCommunicationStrategy 
	extends ObjectStreamCommunicationStrategy {

	public XMLObjectStreamCommunicationStrategy() { }

	public void sendObject(Object destination, Object contents) {
	    super.sendObject(destination, XML.objectToXML(contents));
	}

	public void setupServer(Object destination, MessageListener listener) {
	    final MessageListener innerListener = listener;
	    super.setupServer(destination, new MessageListener() {
		    public void messageReceived(InputMessage message) {
			String xml = (String)message.getContents();
			Object contents = XML.objectFromXML(xml);
			innerListener.messageReceived
			    (new BasicInputMessage(contents));
		    }
		});
	}

    }


    /**
     * A communication strategy in which a destination is mapped to a
     * host and port number, and objects are sent by writing their
     * serialization to a socket.
     */
    public static class ObjectStreamCommunicationStrategy
	implements SocketlikeCommunicationStrategy {

	DestinationTable destinationTable
	    = new BasicDestinationTable(); 	// dest -> data

	HashMap connectionTable
	    = new HashMap();			// dest -> connection

	public ObjectStreamCommunicationStrategy() { }

	/** The data must be a ServiceAddress. */
	public void setDestinationData(Object destination, Object data) {
	    ServiceAddress addr = (ServiceAddress)data;
	    destinationTable.put(destination, addr);
	}

	public Object getDestinationData(Object destination) {
	    Object data = destinationTable.get(destination);
	    if (data == null)
		throw new Error("Unknown IPC destination " + destination);
	    return data;
	}

	public void setDestinationTable(DestinationTable table) {
	    this.destinationTable = table;
	}

	public Connection connectTo(Object destination) {
	    ServiceAddress addr 
		= (ServiceAddress)getDestinationData(destination);
	    Debug.noteln("Connecting to " + destination
			 + " on " + addr.getHost()
			 + " port " + addr.getPort());
	    Debug.assert(connectionTable.get(destination) == null,
			 "connecting twice");
	    try {
		Socket s = new Socket(addr.getHost(), addr.getPort());
		Connection c = new ObjectStreamConnection(s);
		connectionTable.put(destination, c);
		return c;
	    }
	    catch (IOException e) {
		Debug.noteException(e);
		e.fillInStackTrace();
		throw new IPCException(e);
	    }
	}

	public Connection getConnection(Object destination) {
	    Connection c = (Connection)connectionTable.get(destination);
	    if (c == null) {
		c = connectTo(destination);
		connectionTable.put(destination, c);
	    }
	    return c;
	}

	public void sendObject(Object destination, Object contents) {
	    Connection c = getConnection(destination);
	    try { c.send(contents); }
	    catch (IPCException e) {
		discardConnection(destination);
		e.fillInStackTrace();
		throw e;
	    }
	}

	public void discardConnection(Object destination) {
	    Debug.noteln("Discarding connection to", destination);
	    Debug.assert(connectionTable.get(destination) != null);
	    connectionTable.remove(destination);
	}

	public void setupServer(Object destination, MessageListener listener) {
	    // In this method, the destination is the agent who is setting
	    // up the server, not a remote agent.
	    ObjectStreamServer server
		= new ObjectStreamServer(this, destination, listener);
	    server.start();
	}

    }


    /**
     * A Thread that accepts connections to a ServerSocket
     * and creates an object-reading thread for each connection.
     * Each of the object-reading threads notifies the specified
     * listener when an object is received.  The listener's
     * <tt>messageReceived</tt> method should presumably therefore
     * be synchronized.
     */
    public static class ObjectStreamServer extends Thread {

	SocketlikeCommunicationStrategy strategy;
	Object destination;
	MessageListener listener;

	ServiceAddress addr;
	ServerSocket servSock;

	public ObjectStreamServer(SocketlikeCommunicationStrategy strategy,
				  Object destination,
				  MessageListener listener) {
	    this.strategy = strategy;
	    this.destination = destination;
	    this.listener = listener;
	    addr = (ServiceAddress)strategy.getDestinationData(destination);
	}

	public void run() {
	    Debug.noteln("Server running for", destination);
	    try {
		servSock = new ServerSocket(addr.getPort());
		while (true) {
		    Socket s = servSock.accept();
		    Debug.noteln("Client connection", s);
		    receiveFrom(new IPC.ObjectStreamConnection(s));
		}
	    }
	    catch (IOException e) {
		Debug.noteException(e);
		e.fillInStackTrace();
		throw new IPCException(e);
	    }
	}

	protected void receiveFrom(final ObjectStreamConnection connection) {
	    // An exception thrown by connection.receive() will propagate
	    // beyond the run method and hence cause the thread to die,
	    // which is more or less what we want.
	    new Thread() {
		public void run() {
		    Debug.noteln(destination + " receiver running ...");
		    while (true) {
			Debug.noteln(destination + " waiting to receive ...");
			Object contents = connection.receive();
			Debug.noteln(destination + " received", contents);
			listener.messageReceived
			    (new BasicInputMessage(contents));
		    }
		}
	    }.start();
	}
    }


    /**
     * A Connection that is used to send and receive serialized objects
     * via a socket.
     */
    public static class ObjectStreamConnection implements Connection {

	Socket sock;
	ObjectInputStream in;
	ObjectOutputStream out;

	public ObjectStreamConnection(Socket s) /* throws IOException */ {
	    sock = s;
	    /* /\/: For some reason, messages don't seem to get through
	       if we create the Object streams right away. */
	    // in = new ObjectInputStream(sock.getInputStream());
	    // out = new ObjectOutputStream(sock.getOutputStream());
	    // ? do we need to put a buffer in there ?  /\/
	}

	public Socket getSocket() {
	    return sock;
	}

	public void send(Object contents) {
	    try {
		if (out == null)
		    out = new ObjectOutputStream(sock.getOutputStream());
		out.writeObject(contents);
		out.flush();
		// Forget the object(s) just written, so that writing them
		// again will send their full state instead of just a ref.
		out.reset();
	    }
	    catch (IOException e) {
		Debug.noteException(e);
		e.fillInStackTrace();
		throw new IPCException(e);
	    }
	}

	public Object receive() {
	    try {
		if (in == null)
		    in = new ObjectInputStream(sock.getInputStream());
		return in.readObject();
	    }
	    catch (ClassNotFoundException e) {
		Debug.noteException(e);
		e.fillInStackTrace();
		throw new IPCException(e);
	    }
	    catch (IOException e) {
		Debug.noteException(e);
		e.fillInStackTrace();
		throw new IPCException(e);
	    }
	}

    }


    /**
     * An object that contains a host name and a port number.
     */
    public static class ServiceAddress {
	public String host;
	public int port;
	public ServiceAddress(String host, int port) {
	    this.host = host;
	    this.port = port;
	}
	public String getHost() { return host; }
	public int getPort() { return port; }
    }


    /**
     * The exception thrown by IPC methods.  These exceptions are
     * thrown in place of whatever exceptions are thrown internally
     * (such as IOExceptions) in order to hide the internal details.
     */
    public static class IPCException extends RuntimeException {
	Throwable reason;
	public IPCException() {
	    super();
	}
	public IPCException(String message) {
	    super(message);
	}
	public IPCException(Throwable reason) {
	    super("Caused by " + reason.getMessage());
	    this.reason = reason;
	}
	public Throwable getReason() {
	    return reason;
	}
    }


}

// Issues:
// * The name SocketlikeCommunicationStrategy may be too specific;
//   what it really does, we might say, is to expose more of the
//   details.  Connections objects are the most socket-like part;
//   the mapping from destination names to more elaborate "information"
//   objects is much more general.
